/* eslint-disable no-plusplus */
/* eslint-disable class-methods-use-this */
/* eslint-disable no-param-reassign */
import { LitElement, html, css } from 'lit-element';
import '@advanced-rest-client/clipboard-copy/clipboard-copy.js';
import '@advanced-rest-client/json-table/json-table.js';
import '@anypoint-web-components/anypoint-button/anypoint-button.js';
import styles from '@advanced-rest-client/prism-highlight/prism-styles.js';

/** @typedef {import('lit-element').TemplateResult} TemplateResult */
/** @typedef {import('@advanced-rest-client/arc-types').FormTypes.Example} Example */

/**
 * Transforms input into a content to be rendered in the code view.
 */
export const SafeHtmlUtils = {
  AMP_RE: new RegExp(/&/g),
  GT_RE: new RegExp(/>/g),
  LT_RE: new RegExp(/</g),
  SQUOT_RE: new RegExp(/'/g),
  QUOT_RE: new RegExp(/"/g),
  htmlEscape(s) {
    if (typeof s !== 'string') {
      return s;
    }
    if (s.indexOf('&') !== -1) {
      s = s.replace(SafeHtmlUtils.AMP_RE, '&amp;');
    }
    if (s.indexOf('<') !== -1) {
      s = s.replace(SafeHtmlUtils.LT_RE, '&lt;');
    }
    if (s.indexOf('>') !== -1) {
      s = s.replace(SafeHtmlUtils.GT_RE, '&gt;');
    }
    if (s.indexOf('"') !== -1) {
      s = s.replace(SafeHtmlUtils.QUOT_RE, '&quot;');
    }
    if (s.indexOf("'") !== -1) {
      s = s.replace(SafeHtmlUtils.SQUOT_RE, '&#39;');
    }
    return s;
  }
};

/**
 * `api-example-render`
 *
 * Renders a JSON values using Prism highlighter or JSON table.
 *
 * ## Data model
 *
 * The model is generated by `api-example-generator`. Use it to generate
 * view model for examples for AMF shape.
 *
 * ### ExampleModel
 *
 * - **hasRaw** `Boolean` - if true then `raw` property has a value
 * - **hasTitle** `Boolean` - if true then `title` property has a value
 * - **hasUnion** `Boolean` - if true then `values` property has a value
 * - **value** `String`, Optional - Example to render
 * - **title** - `String`, Optional - Example name, only when `hasTitle` is set
 * - **raw** `String`, Optional - Raw value of RAML example. This value is a
 * YAML or JSON schema value. This is only set when raw value is available in
 * the model and it is not JSON/XML.
 * - **values** `Array<ExampleModel>`, Optional - Only when `hasUnion` is set.
 *
 * ## Example
 *
 * ```javascript
 * <api-example-render example="{...}" is-json mime-type="application/json"></api-example-render>
 * ```
 */
export class ApiExampleRender extends LitElement {
  get styles() {
    return [
      styles,
      css`
      :host {
        display: block;
        background-color: inherit;
      }

      .code-wrapper {
        padding: 0px;
      }

      #output {
        white-space: pre-wrap;
        word-wrap: var(--code-block-word-wrap, break-word);
        word-break: var(--code-block-word-break, break-all);
        font-family: var(--arc-font-code-family);
      }

      [hidden] {
        display: none !important;
      }

      .union-toggle {
        outline: none;
        background-color: var(--api-type-document-union-button-background-color, transparent);
        color: var(--api-type-document-union-button-color, #000);
        border-width: 1px;
        border-color: var(--api-body-document-media-button-border-color, #a3b11d);
        border-style: solid;
      }

      .union-toggle[activated] {
        background-color: var(--api-type-document-union-button-active-background-color, #CDDC39);
        color: var(--api-type-document-union-button-active-color, #000);
      }

      .action-button[active] {
        background-color: var(--api-resource-example-document-button-active-background-color, #CDDC39);
        color: var(--api-resource-example-document-button-active-color, currentColor);
      }

      .union-toggle[focused],
      .action-button[active][focused] {
        outline: auto;
      }

      .union-type-selector {
        margin: 0px 8px 12px 0px;
      }

      .code-wrapper.scalar {
        padding-top: 1px;
      }

      .example-actions {
        display: flex;
        align-items: center;
        flex-direction: row;
        justify-content: flex-end;
        flex-wrap: wrap;
        flex: 1;
      }

      anypoint-button {
        margin-bottom: 8px;
        height: 28px;
      }

      api-example-render {
        background-color: inherit;
      }

      json-table,
      api-example-render {
        overflow: auto;
        max-width: 100%;
      }
      `
    ];
  }

  static get properties() {
    return {
      /**
       * Data to render.
       */
      example: { type: Object },
      /**
       * Examples media type
       */
      mediaType: { type: String },
      /**
       * When true the example is a JSON type example.
       */
      isJson: { type: Boolean },
      /**
       * Index of selected union.
       */
      selectedUnion: { type: Number },
      /**
       * Current state of "table" button. When tru the button is highlighted.
       * Note, this won't trigger rendering table/code view as this property is used
       * by `api-resource-example-document` to handle table state change.
       */
      table: { type: Boolean },
      /**
       * When set it renders JSON table instead of code view.
       */
      renderTable: { type: Boolean },
      /**
       * Opens example source view (source from API spec file).
       */
      sourceOpened: { type: Boolean },
      /**
       * When set the actions row (copy, switch view type) is not rendered.
       */
      noActions: { type: Boolean },
      /**
       * Enables Anypoint compatibility styling
       */
      compatibility: { type: Boolean }
    };
  }

  get table() {
    return this._table;
  }

  /**
   * @param {boolean} value
   */
  set table(value) {
    const old = this._table;
    if (old === value) {
      return;
    }
    this._table = value;
    this.requestUpdate('table', old);
    this.dispatchEvent(new CustomEvent('table-changed', {
      composed: true,
      detail: {
        value
      }
    }));
  }

  get mediaType() {
    return this._mediaType;
  }

  /**
   * @param {string} value
   */
  set mediaType(value) {
    const old = this._mediaType;
    if (old === value) {
      return;
    }
    this._mediaType = value;
    this.requestUpdate('mediaType', old);
    this._dataChanged(this._example);
  }

  get example() {
    return this._example;
  }

  /**
   * @param {Example} value
   */
  set example(value) {
    const old = this._example;
    if (old === value) {
      return;
    }
    this._example = value;
    this.requestUpdate('example', old);
    this._dataChanged(value);
    this.selectedUnion = 0;
  }

  get sourceOpened() {
    return this._sourceOpened;
  }

  /**
   * @param {boolean} value
   */
  set sourceOpened(value) {
    const old = this._sourceOpened;
    if (old === value) {
      return;
    }
    this._sourceOpened = value;
    this.requestUpdate('sourceOpened', old);
    this._dataChanged(this._example);
  }

  constructor() {
    super();
    this.sourceOpened = false;
    this.compatibility = false;
    this.table = false;
    this.renderTable = false;
    this.isJson = false;
    this.noActions = false;
    this.sourceOpened = false;
    /** 
     * @type {number}
     */
    this.selectedUnion = undefined;
    this.mediaType = undefined;
  }

  /**
   * Computes whether passed value is a valid JSON object, when component is
   * marked to parse JSON data.
   * @param {boolean} isJson [description]
   * @param {any} value Current example value
   * @return {boolean}
   */
  _computeIsJson(isJson, value) {
    if (!isJson) {
      return false;
    }
    if (!value) {
      return false;
    }
    try {
      const result = JSON.parse(value);
      return typeof result === 'object';
    } catch (_) {
      return false;
    }
  }

  _computeHasRaw(value, raw) {
    if (!raw) {
      return false;
    }
    return String(raw) !== String(value);
  }

  /**
   * @param {Example} example
   */
  _dataChanged(example) {
    if (this.__changeDebouncer || !example) {
      return;
    }
    this.__changeDebouncer = true;
    setTimeout(() => {
      this.__changeDebouncer = false;
      this._renderCode();
    });
  }

  _renderCode() {
    const { example } = this;
    if (!example || (!example.value && example.values)) {
      return;
    }
    const output = /** @type HTMLOutputElement */ (this.shadowRoot.querySelector('#output'));
    if (!output) {
      setTimeout(() => this._renderCode());
      return;
    }
    if (this.sourceOpened) {
      output.innerHTML = this.highlight(String(example.raw), 'yaml');
    } else {
      const value = String(example.value);
      // @ts-ignore
      if (value || value === false || value === 0) {
        output.innerHTML = this.highlight(value, this.mediaType);
      } else {
        output.innerText = '(no value in example)';
      }
    }
  }

  /**
   * Dispatches `syntax-highlight` custom event
   * @param {string} code Code to highlight
   * @param {string} type Mime type of the code
   * @return {string} Highlighted code.
   */
  highlight(code, type) {
    if (code.length > 10000) {
      // examples that are huge causes browser to choke or hang.
      // This just sanitizes the schema and renders unprocessed data.
      return SafeHtmlUtils.htmlEscape(code);
    }
    let lang;
    if (type) {
      if (type.indexOf('json') !== -1) {
        lang = 'json';
      } else if (type.indexOf('xml') !== -1) {
        lang = 'xml';
      }
    }
    const ev = new CustomEvent('syntax-highlight', {
      bubbles: true,
      composed: true,
      detail: {
        code,
        lang
      }
    });
    this.dispatchEvent(ev);
    return ev.detail.code;
  }

  /**
   * Copies the current response text value to clipboard.
   *
   * @param {Event} e
   */
  _copyToClipboard(e) {
    const button = /** @type HTMLButtonElement */ (e.target);
    const copy = this.shadowRoot.querySelector('clipboard-copy');
    if (copy.copy()) {
      button.innerText = 'Done';
    } else {
      button.innerText = 'Error';
    }
    button.disabled = true;
    if ('part' in button) {
      // @ts-ignore
      button.part.add('content-action-button-disabled');
      // @ts-ignore
      button.part.add('code-content-action-button-disabled');
    }
    setTimeout(() => this._resetCopyButtonState(button), 1000);
  }

  /**
   * Resets button icon.
   * @param {HTMLButtonElement} button Button to reset.
   */
  _resetCopyButtonState(button) {
    button.innerText = 'Copy';
    button.disabled = false;
    if ('part' in button) {
      // @ts-ignore
      button.part.remove('content-action-button-disabled');
      // @ts-ignore
      button.part.remove('code-content-action-button-disabled');
    }
    button.focus();
  }

  /**
   * @param {number} selectedUnion
   * @param {Example} example
   * @return {Example|undefined} 
   */
  _computeUnionExamples(selectedUnion, example) {
    if (selectedUnion === undefined || selectedUnion < 0) {
      return undefined;
    }
    if (!example || !example.values) {
      return undefined;
    }
    return example.values[selectedUnion];
  }

  _toggleTable(e) {
    const { target } = e;
    const { value } = e.detail;
    this.table = value;
    if (value && this.sourceOpened) {
      this.sourceOpened = !value;
    }
    this._toggleActionButtonCssPart(target, value);
  }

  _toggleSourceOpened(e) {
    const { target } = e;
    const { value } = e.detail;
    this.sourceOpened = value;
    if (value && this.table) {
      this.table = !value;
    }
    this._toggleActionButtonCssPart(target, value);
  }

  _toggleActionButtonCssPart(target, active) {
    if (!('part' in target)) {
      return;
    }
    const parts = ['content-action-button-active', 'code-content-action-button-active'];
    for (let i = 0, len = parts.length; i < len; i++) {
      if (active) {
        target.part.add(parts[i]);
      } else {
        target.part.remove(parts[i]);
      }
    }
  }

  /**
   * Handler for union type button click.
   * Sets `selectedUnion` property.
   *
   * @param {PointerEvent} e
   */
  _selectUnion(e) {
    const node = /** @type HTMLElement */ (e.target);
    const index = Number(node.dataset.index);
    if (Number.isNaN(index)) {
      return;
    }
    if (this.selectedUnion === index) {
      node.setAttribute('activated', '');
    } else {
      this.selectedUnion = index;
    }
  }

  /**
   * @param {Example} example
   * @returns {TemplateResult|string} 
   */
  _renderUnion(example) {
    const { values } = example;
    if (!values) {
      return '';
    }
    const unions = example.values.map((item) => item.title);
    const {selectedUnion} = this;
    const unionExample = this._computeUnionExamples(selectedUnion, example);
    return html`
      <div class="union-type-selector">
        <span>Any of:</span>
        ${unions.map((item, index) => html`
          <anypoint-button
            class="union-toggle"
            ?activated="${selectedUnion === index}"
            @click="${this._selectUnion}"
            data-index="${index}"
            ?compatibility="${this.compatibility}"
            title="Select ${item} type">${item}</anypoint-button>`)}
      </div>
      ${unionExample ? html`
        <api-example-render
          .example="${unionExample}"
          .isJson="${this.isJson}"
          .mediaType="${this.mediaType}"
          .table="${this.table}"
          .renderTable="${this.renderTable}"
          .noActions="${this.noActions}"></api-example-render>` : ''}
    `;
  }

  /**
   * @param {Example} example
   * @returns {TemplateResult|string} 
   */
  _headerTemplate(example) {
    const noActions = !!(this.noActions || example.isScalar);
    if (noActions) {
      return '';
    }
    const { compatibility } = this;
    const hasRaw = this._computeHasRaw(example.value, example.raw);
    const isJson = this._computeIsJson(this.isJson, example.value);
    return html`
    <div class="example-actions">
      <anypoint-button
        part="content-action-button, code-content-action-button"
        class="action-button"
        data-action="copy"
        @click="${this._copyToClipboard}"
        ?compatibility="${compatibility}"
        title="Copy example to clipboard"
      >Copy</anypoint-button>
      ${isJson ? html`
        <anypoint-button
          part="content-action-button, code-content-action-button"
          class="action-button"
          data-action="table"
          toggles
          .active="${this.table}"
          @active-changed="${this._toggleTable}"
          ?compatibility="${compatibility}"
          title="Toggle between table and JSON view"
        >Table view</anypoint-button>` : ''}
      ${hasRaw ? html`
        <anypoint-button
          part="content-action-button, code-content-action-button"
          class="action-button"
          data-action="code"
          toggles
          .active="${this.sourceOpened}"
          @active-changed="${this._toggleSourceOpened}"
          ?compatibility="${compatibility}"
          title="Toggle between JSON and example source view"
        >Source view</anypoint-button>` : ''}
    </div>`;
  }

  /**
   * @param {Example} example
   * @returns {TemplateResult|string} 
   */
  _renderExample(example) {
    const exampleValue = /** @type any */ (example.value);
    const isJsonExampleValue = this._computeIsJson(this.isJson, exampleValue);
    const renderJsonTable = this.renderTable && isJsonExampleValue;

    return html`
    ${this._headerTemplate(example)}
    ${renderJsonTable ? html`<json-table .json="${exampleValue}"></json-table>`: ''}
    <div class="code-wrapper ${example.isScalar ? 'scalar': ''}" part="code-wrapper, example-code-wrapper" ?hidden="${renderJsonTable}">
      <code id="output" class="markdown-html" part="markdown-html" language-xml=""></code>
    </div>`;
  }

  render() {
    const {example} = this;
    if (!example) {
      return '';
    }
    const isUnion = !!(example && example.hasUnion);
    return html`<style>${this.styles}</style>
    ${isUnion ? this._renderUnion(example) : this._renderExample(example)}
    <clipboard-copy .content="${example.value}"></clipboard-copy>
    `;
  }
}
